/* Copyright (c) 2008, Rob Bloom
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of the <organization> nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY ROB BLOOM ''AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL ROB BLOOM BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>
#include <unistd.h>

#include "linklist.h"

// GLMExtract, pulls files out of Auto Assault's .GLMs files

// Written by Inuchance <inu@inuchance.net>


long int read32(FILE *in);
unsigned char spinner();

int main() {
  printf ("GLM file to eat: ");
  fflush(stdout);
  char *filename;
  filename = malloc(1024);
  fgets(filename,512,stdin);
  int eger;
  for (eger = 0; filename[eger]; eger++) {
    if (filename[eger] == 10 || filename[eger] == 13) {
      filename[eger] = 0;
    }
  }

  FILE *glm;
  glm = fopen(filename, "rb");
  if (!glm) {
    printf("Error opening %s!\n",filename);
    exit(1);
  } else {
    printf("%s opened successfully!\n", filename);
  }
  
  fseek(glm, 0, SEEK_END);
  int glmsize = ftell(glm);

  // This address at the end of the filetells us where CHNKBLXX is at.
  fseek(glm, -4, SEEK_END);
  long int chnkblxx = read32(glm);

  // Jumping to CHNKBLXX to verify it's here.
  fseek(glm,chnkblxx,SEEK_SET);
  char *inp = malloc(32);
  fgets(inp, 9, glm);

  long int filecount;
  if (!strcmp("CHNKBLXX", inp)) {
    printf("CHNKBLXX found! This is a GLM archive.\n");
    long int last = ftell(glm);
    fseek(glm, chnkblxx + 16, SEEK_SET);
    filecount = read32(glm);
    fseek(glm, last, SEEK_SET);
    printf("Contains %li files, archive size is %.3fMB.\n",filecount, (float)glmsize/1024.0/1024.0);
  } else {
    printf("CHNKBLXX not found, therefore, this isn't a GLM archive. :(\n");
    exit(2);
  }

  // Ask to extract, now that we know we will be able to.

  printf("Extract files? [yN] ");
  fflush(stdout);
  fgets(inp, 10, stdin);
  int extract = 0;
  if (tolower(inp[0]) == 'y')
    extract = 1;
    
  LLIST *files = newllist();
  
  // Read all the filenames, then their position and size into a linked list
  long int filenames = read32(glm);
  fseek(glm, filenames, SEEK_SET);
  printf("Reading list of files into memory...\n");
  long int num = 0;
  while (ftell(glm) < chnkblxx) {
    if (!(num%50)) {
      printf ("%5li/%5li read... %c\r", num, filecount, spinner());
      fflush(stdout);
    }
    int filep = 0;
    int c;
    char *x = malloc(128);
    while ((c=fgetc(glm)) != 0) {
      x[filep++] = c;
    }
    x[filep] = 0;
    
    long int lastpos = ftell(glm); // Store pos so we can continue reading filenames afterwards.
    // Let's find the position and length now.
    fseek(glm,chnkblxx + 20 + (0x16 * num), SEEK_SET);
    long int pos = read32(glm);
    long int size = read32(glm);
    //long int wtf = read32(glm); // I don't know what this means... :(
    //long int size2 = read32(glm); // The size again for some reason...
    //long int null = read32(glm); // This seems to always be null...?
    fseek(glm,lastpos,SEEK_SET);
    addlink(files, -1, x, pos, size);
    num++; // Which file is this?  Used for determining offset to size/pos
  }

  printllist(files);

  char *out = malloc(strlen(filename) + 1);
  strcpy(out, filename);
  for (eger = 0; out[eger] != '.' && out[eger] != 0; eger++);
  out[eger] = 0;

  if (extract) {
    printf("Extracting files to %s/...\n",out);
    mkdir(out, S_IRUSR | S_IWUSR | S_IXUSR | S_IXGRP | S_IXOTH | S_IRGRP | S_IROTH);
    if (chdir(out) != 0) {
      printf("Couldn't change to %s/!\n",out);
      exit(3);
    }

    for (eger = 0; eger < files->length; eger++) {
      struct DAT *d = getlink(files,eger);
      printf("Extracting\t: %s...", d->name);
      fflush(stdout);
      FILE *out;
      out = fopen(d->name, "wb");
      fseek(glm, d->glmpos, SEEK_SET);

      char *buf = malloc(d->size);
      if (buf == 0) {
	printf("Not enough memory to read this file into memory! D:\n");
	exit(1);
      }
      int p = 0;
      while (p < d->size)
	buf[p++] = fgetc(glm);
      fwrite(buf, 1, d->size, out);
      free(buf);
      //fputc(fgetc(glm), out);
      
      fclose(out);
      printf("\n");
      
    }
    
  }

  exit(0);
}

long int read32(FILE *in) {
  return
    fgetc(in) + 
    (fgetc(in) << 8) + 
    (fgetc(in) << 16) + 
    (fgetc(in)<<24);
}

const char spinnersize = 4;
const char *spinneranim = "|/-\\";

unsigned char spinner() {
  static unsigned int pos = 0;
  
  if (++pos > spinnersize)
    pos = 0;
  
  return spinneranim[pos];
}
