/* Copyright (c) 2008, Rob Bloom
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of the <organization> nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY ROB BLOOM ''AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL ROB BLOOM BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#include <stdio.h>
#include <stdlib.h>
#include "linklist.h"

// Some generic functions I made to work with linked lists

// Written by Inuchance <inu@inuchance.net>


LLIST *newllist() {
  LLIST *l = malloc(sizeof(LLIST));
  l->length = 0;
  l->cell = NULL;
  return l;
}

void destroyllist(LLIST *l) {
  if (l == NULL)
    return;

  struct DAT *now = l->cell;
  while (l->cell) {
    while (now->next != NULL)
      now = now->next;
    free(now);
  }
  free(l);
  return;
}


void addlink(LLIST *l, int pos, char *name, long int glmpos, long int size) {
  if (l == NULL)
    return;
  if (pos == -1 && l->length == 0)
    pos = 0;
  if (pos < 0)
    pos = l->length + pos + 1;
  if (pos < 0)
    return;

  struct DAT *add = malloc(sizeof(struct DAT));
  add->name = name; add->glmpos = glmpos; add->size = size;

  if (pos == 0) {
    struct DAT *orig = l->cell;
    l->cell = add;
    add->next = orig;
    l->length++; // Keep this count accurate.
    return;
  }

  int p;
  struct DAT *now = l->cell;

  for (p = 1; p < pos && now->next != NULL; p++)
    now = now->next;

  struct DAT *n; // Store the cell's current next d00d
  n = now->next;
  now->next = add; // Now, the next d00d of the cell is our addition
  add->next = n; // and, the next d00d of our addition is the regular cell.  addition to the middle of a list complete!
  l->length++; // Keep this count accurate.

}

void remlink(LLIST *l, int pos) {
  if (l == NULL)
    return;
  if (pos == -1 && l->length == 0)
    pos = 0;
  if (pos < 0)
    pos = l->length + pos + 1;
  if (pos < 0)
    return;
  
  if (pos == 0) { // Removing the first in the list? Need to change the original struct, too, then.
      struct DAT *x = l->cell;
    l->cell = x->next;
    free(x);
    l->length--; // Keep this count accurate.
    return;
  }
  
  int i;
  struct DAT *now = l->cell;
  // Used in case we're deleting cell 1, in which case the 'last' cell (cell before the one being deleted) will point to element 0
  struct DAT *last;
  now = last = l->cell;
  
  for (i = 0; i < pos && now->next != NULL; i++) {
    last = now;
    now = now->next;
  }
  
  if (i == pos) {
    struct DAT *next = now->next;
    last->next = next; // The current (to be removed) cell is dropped from the linked list
      free(now); // And then dropped from system memory
    l->length--; // Keep this count accurate.
  }
}

struct DAT *getlink(LLIST *l, int pos) {
  if (l == NULL)
    return NULL;
  if (pos == -1 && l->length == 0)
    pos = 0;
  if (pos < 0)
    pos = l->length + pos + 1;
  if (pos < 0)
    return NULL;

  if (pos == 0) {
    return l->cell;
  }

  int p;
  struct DAT *now = l->cell;

  for (p = 0; p < pos && now->next != NULL; p++)
    now = now->next;

  return now;
}

void printllist(LLIST *l) {
  if (l == NULL)
    return;

  struct DAT *c;
  int i;
  for (i = 0; i < l->length; i++) {
    c = getlink(l,i);
    printf("pos% 3i:  name=%s, glmpos=0x%08lx, size=%li\n", i,c->name,c->glmpos, c->size);
  }
  printf("list length: %i\n",l->length);
  return;
}
